﻿
#region Using Directives
using RSPhotography.DataAccess.Enums;
using RSPhotography.DataAccess.HelperClasses;
using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Data.Entity.Core.Objects;
using System.Data.Entity.Infrastructure;
using System.Data.Entity.Migrations;
using System.Diagnostics;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
#endregion

namespace RSPhotography.DataAccess.DataAccessors
{
    /// <summary>
    /// This is the base Data Access class
    /// </summary>
    public class DataAccessor<T> : BaseAccessor, IDataAccessor<T>
        where T : EntityBase
    {
        #region Constructors

        /// <summary>
        /// Default constructor - passes to the base class for DbContext instantiation
        /// </summary>
        public DataAccessor() : base()
        {       
        }

        /// <summary>
        /// Sort of like a copy constructor where we reuse an existing DbContext with a new Data Accessor class
        /// </summary>
        /// <param name="existing">A sort of copy constructor where an existing accessor's DbContext is used by this class instance</param>
        protected DataAccessor(BaseAccessor existing)
        {
            var previous = existing as DataAccessor<T>;
            DataContext = previous.DataContext;
            Transaction = previous.Transaction;
        }

        /// <summary>
        /// Sort of like a copy constructor where we reuse an existing DbContext with a new Data Accessor class
        /// </summary>
        /// <param name="access">An existing DbContext</param>
        internal DataAccessor(RSPhotographyEntities access)
        {
            DataContext = access;
        }

        #endregion      

        #region Public Functions

        /// <summary>
        /// A Function which returns multiple matching entitiesof type T
        /// </summary>
        /// <param name="filter">The query filter (criteria)</param>
        /// <param name="navigationProperties">Navigation properties to include from the base entity</param>
        /// <returns>A collection of matching entities</returns>
        public virtual IList<T> GetEntities(Func<T, bool> filter, params Expression<Func<T, object>>[] navigationProperties)
        {   
            IQueryable<T> query = DataContext.Set<T>();

            foreach (Expression<Func<T, object>> navigationProperty in navigationProperties)
            {
                query = query.Include<T, object>(navigationProperty);
            }

            return query.Where(filter).ToList<T>();                        
        }

        /// <summary>
        /// A Function to return a single entity of type T
        /// </summary>
        /// <param name="filter">The query filter (criteria)</param>
        /// <param name="navigationProperties">Navigation properties to include from the base entity</param>
        /// <returns>A matching entity or NULL</returns>
        public virtual T GetEntity(Func<T, bool> filter, params Expression<Func<T, object>>[] navigationProperties)
        {   
            IQueryable<T> query = DataContext.Set<T>();

            foreach (Expression<Func<T, object>> navigationProperty in navigationProperties)
            {
                query = query.Include<T, object>(navigationProperty);
            }

            return query.FirstOrDefault(filter);            
        }

        /// <summary>
        /// Forces pending changes to be committed to the Data Store
        /// </summary>
        public void SaveChanges()
        {
            try
            {
                DataContext.SaveChanges();
            }
            catch (DbUpdateException ex)
            {
                Debug.WriteLine(ex.Message);
                if(ex.InnerException != null)
                {
                    Debug.WriteLine(ex.InnerException.Message);
                }
                throw new ApplicationException();
            }
        }

        /// <summary>
        /// Allows freeform data queries from outside the Data Access classes, without exposing the data context directly
        /// </summary>        
        /// <returns>An IQueryable of type T - limited to 100 rows</returns>
        public IQueryable<T> CreateQuery()
        {
            IQueryable<T> _query = DataContext.Set<T>().AsQueryable<T>();
            _query = _query.Take(100);
            return _query;
        }

        /// <summary>
        /// Allows freeform data queries from outside the Data Access classes, without exposing the data context directly
        /// </summary>  
        /// <param name="maxRows">Max rows to return</param>
        /// <returns>An IQueryable of type T</returns>
        public IQueryable<T> CreateQuery(int maxRows)
        {
            IQueryable<T> _query = DataContext.Set<T>().AsQueryable<T>();

            if (maxRows < 0 || maxRows > 10000)
            {
                maxRows = 1000; //implements an "upper limit" of 1000 rows per query
            }

            _query = _query.Take(maxRows);
            return _query;
        }      

        #endregion

        #region Protected Functions

        /// <summary>
        /// Returns a base query with ** no filters or restrictions ** applied
        /// </summary>
        /// <returns>An IQueryable with no limitations</returns>
        public virtual IQueryable<T> Select()
        {
            return DataContext.Set<T>().AsQueryable<T>();
        }

        /// <summary>
        /// Determines if the specified entity exists in the *local* Db Context cache
        /// Note: does not check the database
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="entity"></param>
        /// <returns></returns>
        protected bool ExistsLocal(T entity)
        {
            return DataContext.Set<T>().Local.Any(e => e == entity);
        }

        /// <summary>
        /// Determines if the specified entity exists in the *local* Db Context cache
        /// Note: does not check the database
        /// </summary>
        /// <typeparam name="T2">The type of the entity</typeparam>
        /// <param name="entity">The entity</param>
        /// <returns></returns>
        protected bool ExistsLocal<T2>(T2 entity) where T2 : EntityBase
        {
            return DataContext.Set<T2>().Local.Any(e => e == entity);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T2"></typeparam>
        /// <param name="entity"></param>
        protected void DetachLocal<T2>(T2 entity) where T2 : EntityBase
        {
            var existing = DataContext.Set<T2>().Local.Any(e => e == entity);
            if (existing)
            {
                DataContext.Set<T2>().Local.Remove(entity);
            }
            DataContext.Set<T2>().Attach(entity);            
        }

        /// <summary>
        /// Determines if an entity exists in the Data Context or in the physical data store
        /// Source (partial): http://stackoverflow.com/questions/6018711/generic-way-to-check-if-entity-exists-in-entity-framework
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="entity"></param>
        /// <returns>Returns null if not found, else returns the located entity which may not be the same as the entity passed in</returns>
        protected T Exists(T entity)
        {
            var objContext = ((IObjectContextAdapter)this.DataContext).ObjectContext;
            var objSet = objContext.CreateObjectSet<T>();
            var entityKey = objContext.CreateEntityKey(objSet.EntitySet.Name, entity);

            DbSet<T> set = DataContext.Set<T>();
            var keys = (from x in entityKey.EntityKeyValues
                        select x.Value).ToArray();

            //Remember, there can by surrogate keys, so don't assume there's just one column/one value
            //If a surrogate key isn't ordered properly, the Set<T>().Find() method will fail, use attributes on the entity to determien the proper order.

            //context.Configuration.AutoDetectChangesEnabled = false;
            //http://stackoverflow.com/questions/11686225/dbset-find-method-ridiculously-slow-compared-to-singleordefault-on-id

            return set.Find(keys);
        }

        /// <summary>
        /// Determines if an entity exists in the Data Context or in the physical data store
        /// Source (partial): http://stackoverflow.com/questions/6018711/generic-way-to-check-if-entity-exists-in-entity-framework
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="entity"></param>
        /// <returns>Returns null if not found, else returns the located entity which may not be the same as the entity passed in</returns>
        protected T2 Exists<T2>(T2 entity) where T2 :  EntityBase
        {
            var objContext = ((IObjectContextAdapter)this.DataContext).ObjectContext;
            var objSet = objContext.CreateObjectSet<T2>();
            var entityKey = objContext.CreateEntityKey(objSet.EntitySet.Name, entity);

            DbSet<T2> set = DataContext.Set<T2>();
            var keys = (from x in entityKey.EntityKeyValues
                        select x.Value).ToArray();

            //Remember, there can by surrogate keys, so don't assume there's just one column/one value
            //If a surrogate key isn't ordered properly, the Set<T>().Find() method will fail, use attributes on the entity to determien the proper order.

            //context.Configuration.AutoDetectChangesEnabled = false;
            //http://stackoverflow.com/questions/11686225/dbset-find-method-ridiculously-slow-compared-to-singleordefault-on-id

            return set.Find(keys);
        }

        /// <summary>
        /// Inserts or updates bulk entities
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="entities"></param>
        public virtual void InsertOrUpdate(params T[] entities)
        {
            ApplyStateChanges(entities);
        }

        /// <summary>
        /// Inserts or Updates based on a defined type
        /// </summary>
        /// <typeparam name="T2">The type of entities</typeparam>
        /// <param name="entities">The entity or entities to process</param>
        public virtual void InsertOrUpdate<T2>(params T2[] entities) where T2 : EntityBase
        {
            ApplyStateChanges<T2>(entities);
        }

        /// <summary>
        /// Allows for the modification of a many-to-many relationship
        /// </summary>
        /// <typeparam name="T2">The type of children entities</typeparam>
        /// <param name="parent">The parent entity</param>
        /// <param name="collectionName">The navigation property name for the relationship to edit</param>
        /// <param name="state">The entity state to change to</param>
        /// <param name="children">A collection of children of type T2</param>
        public void ModifyRelatedEntities<T2>(T parent, string collectionName, EntityState state, params T2[] children)
            where T2 : EntityBase
        {
            DataContext.Set<T>().Attach(parent);
            ObjectContext obj = DataContext.ToObjectContext();
            foreach (T2 child in children)
            {
                DataContext.Set<T2>().Attach(child);
                obj.ObjectStateManager.ChangeRelationshipState(parent, child, collectionName, state);
            }
        }

        /// <summary>
        /// Allows for the modification of a many-to-many relationship
        /// </summary>
        /// <typeparam name="T2">The type of children entities</typeparam>
        /// <param name="parent">The parent entity</param>
        /// <param name="collection">The navigation collection of entity types</param>
        /// <param name="state">The entity state to change to</param>
        /// <param name="children">A collection of children of type T2</param>
        public void ModifyRelatedEntities<T2>(T parent, Expression<Func<T, object>> collection, EntityState state, params T2[] children)       
            where T2 : EntityBase
        {
            try
            {
                DataContext.Set<T>().Attach(parent);
                ObjectContext obj = DataContext.ToObjectContext();
                foreach (T2 child in children)
                {
                    DataContext.Set<T2>().Attach(child);
                    obj.ObjectStateManager.ChangeRelationshipState(parent, child, collection, state);
                }
            }
            catch (NotSupportedException ex)
            {
                Debug.WriteLine(String.Format("An attempt was made to try and modify a relationship which is not supported.  Error: {0}", ex.Message));
            }
        }

        /// <summary>
        /// Sets an entity or entities to be removed on the next Save Changes call
        /// </summary>        
        /// <param name="entities">Entities to delete</param>        
        public virtual void Delete(params T[] entities)
        {
            foreach(T entity in entities.Where(x => x.State != ObjectState.Deleted))
            {
                entity.State = ObjectState.Deleted;
            }
            ApplyStateChanges(entities);
        }

        /// <summary>
        /// Deletes the entities supplied
        /// </summary>
        /// <typeparam name="T2">The type of the entities</typeparam>
        /// <param name="entities">Entities to delete</param>
        public void Delete<T2>(params T2[] entities) where T2 : EntityBase
        {
            foreach(T2 entity in entities.Where(x => x.State != ObjectState.Deleted))
            {
                entity.State = ObjectState.Deleted;
            }
            ApplyStateChanges<T2>(entities);
        }

        #endregion      

        #region Private Functions

        /// <summary>
        /// Processes entities of the parent type
        /// </summary>        
        /// <param name="items">A collection of entities</param>
        private void ApplyStateChanges(params T[] items)
        {
            ApplyStateChanges<T>(items);
        }

        /// <summary>
        /// Loops through entities and determines if they need to be attached, added and then sets the entity state accordingly
        /// </summary>
        /// <typeparam name="T2">Type of items to process</typeparam>
        /// <param name="items">A collection of entities to process</param>
        private void ApplyStateChanges<T2>(params T2[] items) where T2: EntityBase
        {
            Debug.WriteLine("Started ApplyStateChanges");

            DbSet<T2> dbSet = DataContext.Set<T2>();
            
            //ignore anything previously handled
            foreach (T2 item in items.Where(x => x.State != ObjectState.Processed)) 
            {               
                Debug.WriteLine(String.Format("Item: {0}|{1}", item.State, item.GetType()));

                ProcessEntityState<T2>(dbSet, item);                

                foreach (DbEntityEntry<EntityBase> entry in DataContext.ChangeTracker.Entries<EntityBase>()
                    .Where(c => c.Entity.State != ObjectState.Processed && c.Entity.State != ObjectState.Unchanged))
                {
                    Debug.WriteLine(String.Format("Entry: {0}|{1}", entry.Entity.State, entry.Entity.GetType()));

                    var y = DataContext.Entry(entry.Entity);
                    y.State = HelperFunctions.ConvertState(entry.Entity.State);
                    entry.Entity.State = ObjectState.Processed;
                }                
            }
            Debug.WriteLine("Finished ApplyStateChanges");
        }

        /// <summary>
        /// This function checks the state of the entity (of type T2) and handles it accordingly
        /// Note: refactored out of the ApplyStateChanges function, for readability
        /// </summary>
        /// <typeparam name="T2">Type of the parameter arguments</typeparam>
        /// <param name="dbSet">The DataContext set of type (T2)</param>
        /// <param name="item">The item of type (T2) to process</param>
        private void ProcessEntityState<T2>(DbSet<T2> dbSet, T2 item) where T2 : EntityBase
        {
            DbEntityEntry itemEntry = DataContext.Entry<T2>(item);
            // ignore attached entities, only process detached entities
            if (itemEntry.State == EntityState.Detached)
            {
                if (item.State == ObjectState.Added)
                {
                    dbSet.Add(item);
                }
                else
                {
                    if (item.State == ObjectState.Modified)
                    {
                        dbSet.Attach(item);
                    }
                    else if (item.State == ObjectState.Deleted)
                    {
                        //we need to ensure the item actually exists, by loading it explicitly
                        var existing = Exists<T2>(item);
                        if (existing != null)
                        {
                            dbSet.Remove(existing);
                        }
                        //if it doesn't exist, we can't delete it
                    }
                }
            }
        }

        #endregion         
    }
}