﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using RSPhotography.DataAccess;
using System.Diagnostics;
using RSPhotography.DataAccess.DataAccessors;
using System.Linq;
using RSPhotography.DataAccess.HelperClasses;
using System.Data.Entity;
using RSPhotography.DataAccess.Enums;

namespace RSPhotography.UnitTests
{
    [TestClass]
    public class BasicTests
    {
        [TestMethod]
        public void BasicGenericAccessorTest()
        {
            using(DataAccessor<Catalog> acc = new DataAccessor<Catalog>())
            {
                var entity = acc.GetEntity(x => x.CatalogId == 1);
                Assert.IsNotNull(entity, "There should exist an entity with ID = 1");
                var entities = acc.GetEntities(x => x.CatalogId < 10);
                Assert.IsTrue(entities.Count > 0, "Should be at least one entity returned");
            }
        }

        [TestMethod]
        public void BasicGenericAccessorFreeformQuery()
        {
            using (DataAccessor<Catalog> acc = new DataAccessor<Catalog>())
            {
                var entities = acc.CreateQuery().Where(x => x.CatalogId < 10).Include("Genres").ToList();                
                Assert.IsTrue(entities.Count > 0, "Should be at least one entity returned");

                IQueryable<Catalog> extendedQuery = acc.CreateQuery();
                extendedQuery = extendedQuery.Include("Genres");
                extendedQuery = extendedQuery.Where(x => !String.IsNullOrEmpty(x.Title));
                extendedQuery = extendedQuery.Where(x => x.Genres.Count > 1);
                entities = extendedQuery.ToList();
                Assert.IsTrue(entities.Count > 0, "Should be at least one entity returned");
            }
        }

        [TestMethod]
        public void EditM2MRelationship()
        {
            string originalValue = String.Empty;
            string originalBaseValue = String.Empty;
            Genre result = null;
            Catalog editedItem = null;

            using (DataAccessor<Genre> c = new DataAccessor<Genre>())
            {                
                result = c.CreateQuery().Include("Catalogs").Where(x => x.Title == "Landscape").FirstOrDefault();
                Assert.IsNotNull(result, "Should have returned a valid result");
                Assert.IsTrue(result.Catalogs.Count() > 0, "Should be at least one child Catalog");
            }
            using (DataAccessor<Genre> c = new DataAccessor<Genre>())
            {
                editedItem = result.Catalogs.FirstOrDefault();
                editedItem.State = ObjectState.Modified;
                originalValue = editedItem.Title;
                editedItem.Title = "Unit Test Updated";
                originalBaseValue = result.Description;
                result.Description = "New Description";
                result.State = ObjectState.Modified;
                c.InsertOrUpdate(result);
                c.SaveChanges();
            }
            using (DataAccessor<Genre> c = new DataAccessor<Genre>())
            {
                result.Description = originalBaseValue;
                result.State = ObjectState.Modified;
                editedItem.Title = originalValue;
                editedItem.State = ObjectState.Modified;
                c.InsertOrUpdate(result);
                c.SaveChanges();
            }          
        }

        [TestMethod]
        public void BasicDataFetch()
        {
            using (CatalogDataAccessor access = new CatalogDataAccessor())
            {
                var result = access.GetAllItems();
                Assert.IsTrue(result.Count > 0, "Should be at least one record");
                Trace.WriteLine(result[0].Title);
                Catalog first = result[0];
                
                foreach (Location loc in first.Locations)
                {
                    Trace.WriteLine(loc.Title);                    
                }
            }
        }

        [TestMethod]
        public void GetCatalogByRegion()
        {
            using (CatalogDataAccessor access = new CatalogDataAccessor())
            {
                Location loc = new Location();
                loc.LocationId = 1;
                var result = access.FindItemsByLocation(loc);
                Assert.IsTrue(result.Count > 0, "Should be at least one record");
            }
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException), "Expecting the third data access query to fail")]
        public void ObjectTests()
        {
            //new DbContext from the base class
            using (CatalogDataAccessor a = new CatalogDataAccessor()) 
            {
                var result = a.CreateQuery().Where(x => x.CatalogId < 10).ToList();
                using (SizeDataAccessor sa = new SizeDataAccessor(a))
                {
                    var more = sa.CreateQuery().Where(x => x.SizeId < 10).ToList();
                }
                //the following line will throw an exception because the DBContext has been disposed
                var extra = a.CreateQuery().Where(x => x.CatalogId < 10).ToList();
                Assert.IsTrue(extra.Count > 0, "Should be at least one match");
            }
        }

        [TestMethod]
        public void ObjectTestsSuccess()
        {
            CatalogDataAccessor a = new CatalogDataAccessor();
            SizeDataAccessor sa = new SizeDataAccessor(a);

            try
            {
                var result = a.CreateQuery().Where(x => x.CatalogId < 10).ToList();
                var more = sa.CreateQuery().Where(x => x.SizeId < 10).ToList();
            }
            finally
            {
                a.Dispose();
            }
        }

        private static File CreateFile()
        {
            File newFile = new File();
            newFile.Filename = "Unit Test";
            newFile.IsDefault = false;
            newFile.State = ObjectState.Added;
            return newFile;
        }

        private static Size CreateSize()
        {
            Size newSize = new Size();
            newSize.Description = "Unit Test";
            newSize.Dimensions = "U/T";
            newSize.State = ObjectState.Added;
            newSize.SizeId = 500;
            return newSize;
        }

        [TestMethod]
        public void CreateRemoveFK()
        {
            
            File newFile = CreateFile();
            Size newSize = CreateSize();
            using(FileDataAccessor a = new FileDataAccessor())
            {
                a.InsertOrUpdate(newFile);
                a.SaveChanges();
                Assert.IsNotNull(newFile);

                newFile.Size = newSize;
                newFile.SizeId = newSize.SizeId;
                newFile.State = ObjectState.Modified;
            //    newSize.Files.Add(newFile);

                a.InsertOrUpdate(newFile);
                a.SaveChanges();
            }
            using (FileDataAccessor b = new FileDataAccessor())
            {
                File result = b.GetEntity(x => x.FileId == newFile.FileId, x => x.Size);
                Assert.IsNotNull(result);
                Assert.IsNotNull(result.Size);
                newFile = result;
                result.Size = null;
                result.SizeId = null;
                result.State = ObjectState.Modified;
                b.InsertOrUpdate(result);
                b.SaveChanges();
            }

            using (FileDataAccessor c = new FileDataAccessor())
            {
                c.Delete(newSize);
                c.Delete(newFile);
                c.SaveChanges();
            }
        }
    }
}
