﻿using System;
using System.Text;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using RSPhotography.DataAccess.DataAccessors;
using RSPhotography.DataAccess;
using System.Linq;
using System.Data.Entity;
using RSPhotography.DataAccess.Enums;

namespace RSPhotography.UnitTests
{
    [TestClass]
    public class ImportedTests
    {
        [TestMethod]
        public void AddRemoveRelationship()
        {
            Catalog existing = null;

            using (DataAccessor<Catalog> a = new DataAccessor<Catalog>())
            {
                existing = a.CreateQuery().Include("Genres").FirstOrDefault();
                Assert.IsNotNull(existing, "Should find at least one Catalog");
            }

            Genre newEntity = new Genre();
            newEntity.State = ObjectState.Added;
            newEntity.Title = "Unit";
            newEntity.Description = "Test";
            newEntity.GenreId = 1000;

            existing.Genres.Add(newEntity);
            existing.State = ObjectState.Modified;

            using (DataAccessor<Catalog> a = new DataAccessor<Catalog>())
            {
                //breakpoint #1 here
                a.InsertOrUpdate(existing);
                a.SaveChanges();
            }

            newEntity.State = ObjectState.Unchanged;
            existing.State = ObjectState.Modified;

            using (DataAccessor<Catalog> b = new DataAccessor<Catalog>())
            {
                //breakpoint #2 here
                b.ModifyRelatedEntities<Genre>(existing, x => x.Genres, EntityState.Deleted, newEntity);
                b.SaveChanges();
            }

            using (DataAccessor<Genre> c = new DataAccessor<Genre>())
            {
                //breakpoint #3 here
                c.Delete(newEntity);
                c.SaveChanges();
            }
        }

        [TestMethod]
        public void UpdateSingleEntity()
        {
            Catalog existing = null;
            String existingValue = String.Empty;

            using (CatalogDataAccessor a = new CatalogDataAccessor())
            {
                existing = a.CreateQuery().FirstOrDefault();
                Assert.IsNotNull(existing, "Should find at least one Catalog");
                existingValue = existing.Title;
                existing.Title = "Unit " + DateTime.Now.ToString("MMdd hh:mm:ss");
            }
            using (CatalogDataAccessor b = new CatalogDataAccessor())
            {
                existing.State = ObjectState.Modified;
                b.InsertOrUpdate(existing);
                b.SaveChanges();
            }
            using (CatalogDataAccessor c = new CatalogDataAccessor())
            {
                existing.Title = existingValue;
                existing.State = ObjectState.Modified;
                c.InsertOrUpdate(existing);
                c.SaveChanges();
                Assert.IsTrue(c.CreateQuery().Where(x => x.Title == existingValue).FirstOrDefault() != null, "Update should have worked");
            }
        }

        [TestMethod]
        public void UpdateSingleEntitySingleContext()
        {
            Catalog existing = null;
            String existingValue = String.Empty;

            using (CatalogDataAccessor a = new CatalogDataAccessor())
            {
                existing = a.CreateQuery().FirstOrDefault();
                Assert.IsNotNull(existing);
                existingValue = existing.Title;
                existing.Title = "Unit " + DateTime.Now.ToString("MMdd hh:mm:ss");

                existing.State = ObjectState.Modified;
                a.InsertOrUpdate(existing);

                existing.Title = existingValue;
                existing.State = ObjectState.Modified;
                a.InsertOrUpdate(existing);
                Assert.IsTrue(a.CreateQuery().Where(x => x.Title == existingValue).FirstOrDefault() != null, "Update should have worked");
            }
        }        

        [TestMethod]
        public void UpdateManyToMany()
        {
            Catalog existing = null;
            Genre other = null;
            String existingValue = String.Empty;
            String existingOtherValue = String.Empty;

            using (CatalogDataAccessor a = new CatalogDataAccessor())
            {
                //Note that we include the navigation property in the query
                existing = a.CreateQuery().Include("Genres").FirstOrDefault();
                Assert.IsTrue(existing.Genres.Count() > 1, "Should be at least 1 linked item");
            }
            //save the original description
            existingValue = existing.Description;
            //set a new dummy value (with a date/time so we can see it working)
            existing.Description = "Edit " + DateTime.Now.ToString("yyyyMMdd hh:mm:ss");
            existing.State = ObjectState.Modified;

            other = existing.Genres.First();
            //save the original value
            existingOtherValue = other.Description;
            //set a new value
            other.Description = "Edit " + DateTime.Now.ToString("yyyyMMdd hh:mm:ss");
            other.State = ObjectState.Modified;

            //a new data access class (new DbContext)
            using (CatalogDataAccessor b = new CatalogDataAccessor())
            {
                //single method to handle inserts and updates
                b.InsertOrUpdate(existing);
            }

            //return the values to the original ones
            existing.Description = existingValue;
            other.Description = existingOtherValue;
            existing.State = ObjectState.Modified;
            other.State = ObjectState.Modified;

            using (CatalogDataAccessor c = new CatalogDataAccessor())
            {
                //update the entities back to normal
                c.InsertOrUpdate(existing);
            }
        }

        [TestMethod]
        public void UpdateManyToManySingleContext()
        {
            Catalog existing = null;
            Genre other = null;
            String existingValue = String.Empty;
            String existingOtherValue = String.Empty;

            using (CatalogDataAccessor a = new CatalogDataAccessor())
            {
                existing = a.CreateQuery().Include("Genres").Where(x => x.Genres.Count() > 0).FirstOrDefault();
                Assert.IsTrue(existing.Genres.Count() > 0, "Should be at least 1 linked item");

                existingValue = existing.Description;
                existing.Description = "Edit " + DateTime.Now.ToString("yyyyMMdd hh:mm:ss");
                existing.State = ObjectState.Modified;

                other = existing.Genres.First();
                existingOtherValue = other.Description;
                other.Description = "Edit " + DateTime.Now.ToString("yyyyMMdd hh:mm:ss");
                other.State = ObjectState.Modified;

                a.InsertOrUpdate(existing);

                existing.Description = existingValue;
                other.Description = existingOtherValue;
                existing.State = ObjectState.Modified;
                other.State = ObjectState.Modified;

                a.InsertOrUpdate(existing);
            }
        }

        [TestMethod]
        public void InsertManyToManyThenDelete()
        {
            Catalog existing = null;

            using (CatalogDataAccessor accessor = new CatalogDataAccessor())
            {
                existing = accessor.CreateQuery().Include("Genres").FirstOrDefault();
                Assert.IsNotNull(existing, "Should have at least oen record");              
            }

            File newFile = new File();
            newFile.Filename = "Unit Test";
            newFile.IsDefault = false;
            newFile.State = ObjectState.Added;
            newFile.Catalog = existing;
            newFile.CatalogId = existing.CatalogId;

            existing.Files.Add(newFile);
            existing.State = ObjectState.Modified;

            using (CatalogDataAccessor b = new CatalogDataAccessor())
            {
                b.InsertOrUpdate(existing);
                b.SaveChanges();
            }

            Size newSize = new Size();
            newSize.Description = "UT " + DateTime.Now.ToString("yyyyMMdd hh:mm:ss");
            newSize.SizeId = 1000;
            newSize.Dimensions = "UT";
            newSize.State = ObjectState.Added;

            newFile.Size = newSize;
            newFile.SizeId = newSize.SizeId;
            newFile.State = ObjectState.Modified;

            using (SizeDataAccessor b2 = new SizeDataAccessor())
            {
                b2.InsertOrUpdate(newSize);
                b2.SaveChanges();
            }

            using (FileDataAccessor c = new FileDataAccessor())
            {
                newFile.CatalogId = null;
                newFile.Catalog = null;
                c.InsertOrUpdate(newFile);
                c.SaveChanges();
            }

            using (DataAccessor<EntityBase> d = new DataAccessor<EntityBase>())
            {
                newFile.Size = null;
                newFile.SizeId = null;     
                newFile.State = ObjectState.Modified;
                newSize.State = ObjectState.Modified;                
                newSize.Files.Remove(newFile);

                d.InsertOrUpdate(newFile);
                d.SaveChanges();
                
                newFile.State = ObjectState.Deleted;
                newSize.State = ObjectState.Deleted;                
                
                d.Delete<Size>(newSize);
                d.Delete<File>(newFile);
                
                d.SaveChanges();
            }
        }

        [TestMethod]
        public void InsertManyToManyThenDeleteSingleContext()
        {
            Catalog existing = null;

            using (CatalogDataAccessor accessor = new CatalogDataAccessor())
            {
                existing = accessor.CreateQuery().Include("Genres").FirstOrDefault();
                Assert.IsNotNull(existing, "Should have at least one record");

                File newFile = new File();
                newFile.Filename = "Unit Test";
                newFile.IsDefault = false;
                newFile.State = ObjectState.Added;
                newFile.Catalog = existing;
                newFile.CatalogId = existing.CatalogId;

                existing.Files.Add(newFile);
                existing.State = ObjectState.Modified;

                accessor.InsertOrUpdate(existing);
                accessor.SaveChanges();

                Size newSize = new Size();
                newSize.Description = "UT " + DateTime.Now.ToString("yyyyMMdd hh:mm:ss");
                newSize.SizeId = 1000;
                newSize.Dimensions = "UT";
                newSize.State = ObjectState.Added;

                newFile.Size = newSize;
                newFile.SizeId = newSize.SizeId;
                newFile.State = ObjectState.Modified;

                accessor.InsertOrUpdate<Size>(newSize);
                accessor.SaveChanges();

                newFile.CatalogId = null;
                newFile.Catalog = null;
                accessor.InsertOrUpdate<File>(newFile);
                accessor.SaveChanges();

                newFile.Size = null;
                newFile.SizeId = null;
                newFile.State = ObjectState.Modified;
                newSize.State = ObjectState.Modified;
                newSize.Files.Remove(newFile);

                accessor.InsertOrUpdate<File>(newFile);
                accessor.SaveChanges();

                newFile.State = ObjectState.Deleted;
                newSize.State = ObjectState.Deleted;

                accessor.Delete<Size>(newSize);
                accessor.Delete<File>(newFile);

                accessor.SaveChanges();
            }
        }
    }   
}
